#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

#############################################################################
# PgcEdit.tcl Copyright 2004, 2005 r0lZ
# Program main entry.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

#############################################################################
#
# PgcEdit requires Tcl/Tk 8.4 or greather
#
#############################################################################

set ::program_version 0.6.2
set ::release_date "December 2, 2005"
set ::config(debug) 0
set ::config(expert_mode) false

wm withdraw .
# necessary for freeWrap 5.61
if {!$tcl_interactive} {
	catch {console hide}
}
update idletasks

#catch {encoding system cp1252}
#catch {encoding system utf-8}
catch {encoding system identity}

if {$tcl_version < 8.4} {
	package require Tk
	tk_messageBox -icon error -title "PgcEdit $::program_version error" -parent . \
		-message "PgcEdit requires Tcl/Tk version 8.4 or greather.\nYou are currently running Tcl/Tk version $tcl_patchLevel.\nSorry."
	exit 1000
}

package require Tk 8.4

#############################################################################
## Debug print.
## Usage: set ::config(debug) <triggerlevel>
##        dprint <level> <message>

proc dprint {l t} {
	catch {if {$::config(debug)>=$l} {puts $t}}
}


#############################################################################
## show_msg
## Show message in the status line
proc show_msg {msg {pos 0} {max 100}} {
	set txt [string map {"\n" "  "} $msg]
	.mf.st.f.status itemconfigure $::progressbar_text -text $txt
	.mf.st.f.status configure -width [expr [font measure $::config(global_font) $txt] + 2]
	update_progress $pos $max
	return
}
proc update_progress {pos {max 100}} {
	if {$pos >= 0 && [info exists ::progressbar_item]} {
		.mf.st.f.status delete $::progressbar_item
		unset ::progressbar_item
	}
	if {$pos > 0} {
		set ::progressbar_item [.mf.st.f.status create rectangle 0 0 \
			[expr {round(([winfo width .mf.st.f.status]-5)*((1.0*$pos)/$max))}] \
			30 -fill lightYellow -outline lightYellow]
		.mf.st.f.status raise $::progressbar_text
	}
	update idletasks
	set ::pause_done 0
	after 1 {set ::pause_done 1}
	tkwait variable ::pause_done
}

# Test PgcEdit version.
# Return true if current PgcEdit version >= minversion
# This check takes "beta#" into account! (ie 0.6.1beta1 = 0.6.1)
# If the test fails, an error message is displayed.
# An empty application name skips the message (the default)
# Example: test_version 0.6.1.3
# Example: test_version 0.6.1 "PUO plugin"
proc test_version {minversion {appname ""}} {
	set minlst [split [string map {beta .} $minversion.1000] .]
	set curlst [split [string map {beta .} $::program_version.1000] .]
	if {[lindex $minlst 0] < [lindex $curlst 0]} {set rc true}
	if {[lindex $minlst 0] > [lindex $curlst 0]} {set rc false}
	if {![info exists rc]} {
		if {[lindex $minlst 1] < [lindex $curlst 1]} {set rc true}
		if {[lindex $minlst 1] > [lindex $curlst 1]} {set rc false}
		if {![info exists rc]} {
			if {[lindex $minlst 2] < [lindex $curlst 2]} {set rc true}
			if {[lindex $minlst 2] > [lindex $curlst 2]} {set rc false}
			if {![info exists rc]} {
				if {[lindex $minlst 3] <= [lindex $curlst 3]} {
					set rc true
				} else {
					set rc false
				}
			}
		}
	}
	if {!$rc && $appname != ""} {
		set txt $appname:
		set txt "$txt\n\nThis feature requires PgcEdit version $minversion or more."
		set txt "$txt\nYou are currently running PgcEdit version $::program_version."
		set txt "$txt\nPlease upgrade PgcEdit."
		tk_messageBox -title "PgcEdit: $appname" -icon error -message $txt
	}
	return $rc
}

#############################################################################
## show_warning
## Show message in a message box and in the status line.
## Usage: show_warning -message "message" [-title "title"] [-icon info|warning|error] [-parent window]
proc show_warning {args} {
	set title "PgcEdit"
	set icon "info"
	set parent .

	set idx 0
	set opt [lindex $args $idx]
	while {$opt != ""} {
		incr idx
		set val [lindex $args $idx]
		incr idx
		switch -- $opt {
			"-message" {set msg $val}
			"-title" {set title $val}
			"-icon" {set icon $val}
			"-parent" {set parent $val}
		}
		set opt [lindex $args $idx]
	}
	show_msg $msg
	if {$msg != ""} {
		if { ! $::config(expert_mode) || $icon == "error"} {
			tk_messageBox -message $msg -title $title -icon $icon -parent .
		}
	}
}

proc tk_chooseNewDirectory {args} {
	if {$::config(use_file_browser)} {
		set destfile [eval tk_getSaveFile -initialfile "VIDEO_TS.IFO" \
				-filetypes \{\{"All files" \{*.*\}\} \{"DVD files" \{*.IFO *.BUP *.VOB\}\} \{"VIDEO_TS.IFO" \{VIDEO_TS.IFO\}\}\} \
				$args]
		if {$destfile == ""} {return ""}
		set destdir [file dirname $destfile]
	} else {
		set destdir [eval tk_chooseDirectory $args -mustexist false]
		if {$destdir == ""} {return ""}
	}

	if {! [file exists $destdir]} {
		set pathname $destdir
		while {![file exists $pathname]} {
			set newdir [file dirname $pathname]
			if {$pathname == $newdir} {
				set pathname ""
			} else {
				set pathname $newdir
			}
		}
		set lastpart [file tail $pathname]
		set newparts [string range $destdir [expr {[string length $pathname]+1}] end]
		set newpart1 [lindex [file split $newparts] 0]
		if {$lastpart == $newpart1} {
			set newpartslist [file split $newparts]
			for {set i 1} {$i < [llength $newpartslist]} {incr i} {
				set pathname [file join $pathname [lindex $newpartslist $i]]
			}
		} else {
			set pathname $destdir
		}

		set r [tk_messageBox -title "Browse for Folder" -icon question \
				-message "Folder \"[file nativename $pathname]\\\" does not exist.\nCreate it ?" \
				-parent . -type yesno -default yes]
		if {$r == "no"} {
			return ""
		} elseif {[makedir $pathname]} {
			return ""
		}
	} else {
		set pathname $destdir
	}

	# try to avoid the 'transparent click' Tk bug
	update idletasks
	after 100 {set ::tk_choose_dir_delay 1}
	tkwait variable ::tk_choose_dir_delay

	return $pathname
}

proc tk_chooseOldDirectory {args} {
	if {$::config(use_file_browser)} {
		set destfile [eval tk_getOpenFile -initialfile "VIDEO_TS.IFO" \
				-filetypes \{\{"All files" \{*.*\}\} \{"DVD files" \{*.IFO *.BUP *.VOB\}\} \{"VIDEO_TS.IFO" \{VIDEO_TS.IFO\}\}\} \
				-defaultextension 2 $args]
		if {$destfile == ""} {return ""}
		set destdir [file dirname $destfile]
	} else {
		set destdir [eval tk_chooseDirectory $args -mustexist true]
	}
	
	# try to avoid the 'transparent click' Tk bug
	update idletasks
	after 100 {set ::tk_choose_dir_delay 1}
	tkwait variable ::tk_choose_dir_delay

	return $destdir
}

#############################################################################
## Procedure:  init_user_prefs
## Init user modifiables constants
namespace eval ::tr {}

proc init_user_prefs {} {
	global argv0
	# New to v0.3: config files are stored in "~/Application Data/PgcEdit/"
	# instead of "~/", under Windows.

	if {$::tcl_platform(platform) == "windows"} {
		set ::config(config_directory) [file join $::env(HOME) "PgcEdit"]
		# $::env(HOME) is wrong on some foreign Windows systems
		if {! [catch {set tmp [registry get "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders" "AppData"]} err_str]} {
			set ::config(config_directory) [file join $tmp "PgcEdit"]
		} elseif {[file exists [file join $::env(HOME) "Application Data"]]} {
			set ::config(config_directory) [file join $::env(HOME) "Application Data/PgcEdit"]
		}
	} else {
		set ::config(config_directory) [file join $::env(HOME) ".PgcEdit"]
	}

	if {![file exists $::config(config_directory)]} {
		set old_config_directory $::env(HOME)
		if {$::tcl_platform(platform) == "windows"} {
			if {[file exists [file join $::env(HOME) "Application Data"]]} {
				set old_config_directory [file join $::env(HOME) "Application Data/PgcEdit"]
			}
			# $::env(HOME) is wrong on some foreign Windows systems
			if {[info exists ::freewrap::progname] && \
					[info script] == "" && ! [file exists [file join $::env(HOME) "Application Data"]]} {
				if {! [catch {set tmp [registry get "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders" "AppData"]} err_str]} {
					set old_config_directory [file join $tmp "PgcEdit"]
				}
			}
		} else {
			set old_config_directory [file join $::env(HOME) ".PgcEdit"]
		}
		makedir $::config(config_directory)
		if {[file exists [file join $old_config_directory .pgcedit.cfg]]} {
			foreach i {.pgcedit.cfg .pgcedit.tools .burn.cfg .pgcedit.player.cfg .pgcedit.scratchbook.sbk preview.cfg.ini pgcedit.user.tcl} {
				catch {file rename [file join $old_config_directory $i] $::config(config_directory)}
				catch {file rename [file join $::env(HOME) $i] $::config(config_directory)}
			}
		}
	}
	# remove leading dot
	if {![file exists [file join $::config(config_directory) "pgcedit.cfg"]]} {
		foreach i {pgcedit.cfg pgcedit.tools burn.cfg pgcedit.player.cfg pgcedit.scratchbook.sbk} {
			catch {file rename [file join $::config(config_directory) ".$i"] [file join $::config(config_directory) $i]}
		}
	}

	set ::config(restore_defaults) false

	if {$::tcl_platform(platform) == "windows"} {
		set ::config(tk) win
		set ::config(platform) win
		set rootdir "C:"
		foreach i {C: D: E: F: G: H: I: J: K: L: M: N: O: P: Q: R: S: T: U: V: W: X: Y: Z:} {
		if {[file exists $i]} {
				set rootdir "$i\\"
				break
			}
		}
		if {! [catch {set tmp [registry get "HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion" "ProgramFilesDir"]} err_str]} {
			set progfiles "$tmp\\"
		} else {
			set progfiles "${rootdir}Program files\\"
		}
	} else {
		set ::config(tk) x11
		if {$::tcl_platform(platform) == "macintosh" || \
			[string first "MAC" [string toupper $::tcl_platform(machine)]] >= 0 ||
			[string first "DARWIN" [string toupper $::tcl_platform(machine)]] >= 0} {
			set ::config(platform) mac
		} else {
			set ::config(platform) linux
		}

		# cheap way to test if we are running under Aqua
		button .b
		set tmp [.b cget -padx]
		if {[string is integer $tmp] && $tmp >= 11} {
			set ::config(tk) aqua
		}
		destroy .b

		set rootdir {/}
		set progfiles {/}
	}

	# Automatically save setups on program exit?
	set ::config(save_setups_on_exit) true

	# fonts
	label .testfont
	set ::config(global_font) [.testfont cget -font]
	destroy .testfont
	if {$::tcl_platform(platform) == "windows"} {
		set ::config(fixed_width_font) {{Courier} 9}
	} else {
		set ::config(fixed_width_font) {{Courier} 8}
		set ::config(global_font) [list [lindex $::config(global_font) 0] 8]
	}

	# Tearoff menus
	set ::config(tearoff_menus) true

	# Automatically launch Open DVD dialog at startup?
	set ::config(auto_open_new_dvd) false

	# Automatically open trace mode when a DVD is opened?
	set ::config(open_trace_mode) false

	# Use file browser instead of folder browser to Open/Save DVD?
	set ::config(use_file_browser) false

	# Automatically create backup when a DVD is opened?
	set ::config(auto_create_backup) true
	
	# Automatically launch the "Fix number of streams" function when a DVD is opened?
	set ::config(auto_fix_streams) true
	
	# Put backup dir in DVD's root dir or in VIDEO_TS
	set ::config(backup_in_root) false

	# When saving the DVD, ensure that there is at least 32K of space between
	# end of IFO and start of BUP?  (also used to calculate the DVD space)
	set ::config(save_BUP_with_gaps) false

	# Show the PGC selector at startup?
	set ::config(selector) true

	# which side to display the selector?
	set ::config(selector_side) left

	# height of the selector (for top and bottom sides only)?
	set ::config(selector_height) 8

	# Show the action buttons at startup?
	set ::config(buttons) true

	# Show the toolbar at startup?
	set ::config(toolbar) true
	
	# Automatically show the action buttons at startup?
	set ::config(searchstring) ""

	# Automatically adjust Command Editor window size?
	set ::config(adjust_win_size) true

	# Default colors for GUI
	if {$::config(tk) == "x11"} {
		set ::config(use_custom_color) true
		set ::config(custom_color) #F0F0F0
	} else {
		set ::config(use_custom_color) false
		set ::config(custom_color) [[button .b] cget -background]; destroy .b
	}
	set ::config(white_listbox_background) true

	# Renumber GOTSs?
	set ::config(renumber_gotos) false

	# Highlight errors and Goto targets?
	set ::config(highlight_gotos) true

	# Display [hex codes] in main window?
	set ::config(show_hex_codes) false

	# Display [hex codes] in Info -> Dumps?
	set ::config(dump_hex_codes) false

	# Allow automatic comments in PGC Selector pane?
	set ::config(allow_auto_comments) true

	# Automatically load/save user comments in backup folder?
	set ::config(auto_backup_comments) true

	# Expert mode?
	set ::config(expert_mode) false

	# menubuttons parser: launch parser at DVD load?
	set ::config(auto_launch_menubuttons_parser) true

	# menubuttons parser: default number of LBAs to skip in title domains
	set ::config(lbas_to_skip) 4

	# menubuttons parser: default number of LBAs to skip in title domains
	set ::config(check_multiple_BOVs_sets) false

	# The startup directory for the DVD file browser:
	set ::config(initialdir) $rootdir

	# The startup directory for the Load Menu file browser:
	set ::config(importmenuinitialdir) $rootdir

	# The startup directory for the scratchbook file browser:
	set ::config(scratchbook_initialfile) $rootdir

	# The startup directory for the scratchbook file browser:
	set ::config(import_initialdir) $rootdir

	# The startup filename for the dump file browser:
	set ::config(dumpinitialfile) "${rootdir}dvd_dump.txt"

	# Size and position of the main window
	set ::config(main_win_geometry) 850x500

	# Position of the Command Editor window
	set ::config(ced_win_geometry) ""

	# Position of the Trace Info window
	set ::config(trace_info_win_geometry) ""

	# Position of the Menu Buttons Info window
	set ::config(menubuttons_win_geometry) ""

	# Position of the Menu Editor window
	set ::config(menueditor_win_geometry) ""

	# Position of the Menu Color schemes window
	set ::config(colorschemes_win_geometry) ""

	# Use preview?  (Set in code, before calling this function)

	# Divisor for the Menu Buttons Video Screen window
	set ::config(preview_screen_divisor) 1.333333

	# Trace: highlight last step in log?
	set ::tr::hide_gui true

	# Trace window always on top (yes=".", no="")
	set ::tr::watch_always_on_top "."

	# Trace: highlight last step in log?
	set ::tr::highlight_log_step true

	# Trace Info View flags
	set ::tr::view_gprms 1
	set ::tr::view_sprms 1
	set ::tr::view_puos  0
	set ::tr::view_input 1
	set ::tr::view_slabs 1

	# Automatically start playback of the preview?
	set ::config(preview_autoplay) 0

	# Disable DirectDraw Overlays?
	if {$::tcl_platform(platform) == "windows"} {
		set ::config(preview_no_overlay) 0
	} else {
		set ::config(preview_no_overlay) 1
	}

	# launch preview automatically when playing a cell in trace mode?
	set ::config(trace_auto_preview) false
	
	# skip preview of small blanked cells in trace mode?
	set ::config(trace_skip_small_cells_preview) false

	# close preview automatically at end ot cell in trace mode?
	set ::config(trace_auto_close) 0

	# menu editor outline colors
	set ::config(menued_outline_color) cyan
	set ::config(menued_phantom_color) green
	set ::config(menued_other_buttons_color) #D00000
	# menu editor show links
	set ::config(menued_show_all_buttons) true
	# menu editor show links
	set ::config(menued_show_links) false
	# menu editor show advanced controls panel
	set ::config(menued_advanced_controls) false

	# close preview automatically at end ot cell in trace mode?
	set ::tr::delay 5.0

	# get a web browser.
	set ::config(web_browser) ""
	if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname]} {
		catch {set ::config(web_browser) \
				[lindex [string map {\\ /} [::freewrap::shell_getCmd_imp .html open]] 0] }
	}
	if {$::config(platform) == "mac" && $::tcl_platform(platform) == "unix"} {
		set ::config(web_browser) "open"
	}

	# Open popup menus when right mouse button is released?  (not recommended)
	set ::config(popups_button) "<3>"


	### Not available via GUI:

	# Number of sectors for AutoLocate function of Preview
	set ::config(preview_autolocate) 500


	# v0.1.3 bug in config file workaround
	set ::config(selector_win_geometry) ""

	# build default scratchbook filename
	set ::config(default_scratchbook_filename) \
			[file join $::config(config_directory) pgcedit.scratchbook.sbk]

	# Burn function parameters
	namespace eval ::burn {}
	if {$::tcl_platform(platform) == "windows"} {
		set ::burn::mkisofspath "${progfiles}ImgTool Classic\\mkisofs.exe"
		set ::burn::imgburnpath "${progfiles}ImgBurn\\ImgBurn.exe"
		set ::burn::daemonpath "${progfiles}D-Tools\\daemon.exe"
		set ::burn::out "${rootdir}\\DVDimage.ISO"
	} else {
		set ::burn::mkisofspath "mkisofs"
		set ::burn::imgburnpath "n/a"
		set ::burn::daemonpath "n/a"
		set ::burn::out "DVDimage.ISO"
	}
	if {$::config(platform) == "mac"} {
		set tmp [file dirname [file dirname [file normalize [info script]]]]
		foreach i {{} { x11} { X11} { aqua} { Aqua} {  AQUA} {  x11} {  X11} {  aqua} {  Aqua} {  AQUA}} {
			if {[file exists [file join $tmp "PgcEdit Starter$i.app/Contents/Resources/bin/mkisofs"]]} {
				set ::burn::mkisofspath [file join $tmp "PgcEdit Starter$i.app/Contents/Resources/bin/mkisofs"]
				break
			}
		}
	}

	set ::burn::splitoutput {}
	set ::burn::showlog true
	set ::burn::burnerletter {}
	set ::burn::sldisksize 2295104	;# 2295104 for a DVD+R, or approx 2297888 for a DVD-R
	set ::burn::dldisksize 4173824	;# 4173824 for a DL DVD+R, or minimum 4150400 for a DL DVD-R
	set ::burn::layer_break_sector 2092896	;# required for burning a DL DVD-R
	set ::burn::mount false
	set ::burn::mount_device 0
	set ::burn::burn false
	set ::burn::speed MAX
	set ::burn::eraserw ""
	set ::burn::verify NO
	set ::burn::start /START
	set ::burn::close ""
	set ::burn::shutdown ""
	set ::burn::deleteimage NO
	# the first 4 options are saved separately in the burn.cfg file.
	set configfile [file join $::config(config_directory) "burn.cfg"]
	if {[file exists $configfile]} {
		if {[catch {source $configfile} err] != 0} {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "Error in config file $configfile:\n$err" -parent .
		} else {
			if {![info exists ::burn::media]} {
				if {$::burn::layer_break_sector == 0} {
					set ::burn::layer_break_sector 2092896
					set ::burn::media "Double layer DVD+R"
				} else {
					set ::burn::media "Dual layer DVD-R"
				}
			}
			if {[info exists ::burn::dvddecrypterpath]} {
				set ::burn::imgburnpath $::burn::dvddecrypterpath
				if {!$::config(illegal_aps)} {
					set txt "DVD Decrypter is now discontinued.\n"
					set txt "${txt}It has been replaced by ImgBurn.\n\n"
					set txt "${txt}ImgBurn is also programmed by Lightning UK!, is also free,\n"
					set txt "${txt}and has exactly the same burn capabilities than DVD\n"
					set txt "${txt}Decrypter, plus some improvments (including a command\n"
					set txt "${txt}line option to pass the layer break sector number, which\n"
					set txt "${txt}is now used by PgcEdit when burning DL discs, instead of\n"
					set txt "${txt}the registry method used for DVD Decrypter.)\n\n"
					set txt "${txt}Although it is still possible to use DVD Decrypter to burn\n"
					set txt "${txt}a DVD from PgcEdit, it is recommended to download and\n"
					set txt "${txt}use ImgBurn instead.\n"
					set txt "${txt}Please go to the Burn Setup dialog to change your config.\n\n"
					set txt "${txt}This message will be shown until you save the burn setups\n"
					set txt "${txt}(with either ImgBurn.exe or DVDDecrypter.exe configured\n"
					set txt "${txt}in the ImgBurn path field.)"
					tk_messageBox -message $txt -icon info -title "PgcEdit: Important note"
				}
			}
		}
	} else {
		set ::burn::media "Double layer DVD+R"
	}

	# on Mac systems, use some better defaults
	if {$::config(tk) == "aqua"} {
		set ::config(tearoff_menus) false	;# bug with some versions of Tk
		if {$::tcl_platform(platform) == "macintosh"} {
			set ::config(global_font) {{Times} 12}
		} else {
			set ::config(global_font) {{Lucida Grande} 12}
		}
		set ::config(fixed_width_font) {Monaco 11}
		set ::config(main_win_geometry) 877x539
		set ::config(ced_win_geometry) 464x270
		set ::config(adjust_win_size) true
	} elseif {$::config(tk) == "x11"} {
		set ::config(global_font) {gothic 11}
		set ::config(fixed_width_font) {fixed 12}
	}

	#################################
	# Overwrite default values with user defined ones
	set configfile [file join $::config(config_directory) "pgcedit.cfg"]
	if {[file exists $configfile]} {
		if {[catch {source $configfile} err] != 0} {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}

	# if ::config(initialdir) does not exists, find the first directory that exists
	while {![file exists $::config(initialdir)]} {
		set newdir [file dirname $::config(initialdir)]
		if {$::config(initialdir) == $newdir} {
			set ::config(initialdir) $rootdir
		} else {
			set ::config(initialdir) $newdir
		}
	}

	# if ::config(importmenuinitialdir) does not exists, find the first directory that exists
	while {![file exists $::config(importmenuinitialdir)]} {
		set newdir [file dirname $::config(importmenuinitialdir)]
		if {$::config(importmenuinitialdir) == $newdir} {
			set ::config(importmenuinitialdir) $rootdir
		} else {
			set ::config(importmenuinitialdir) $newdir
		}
	}

	# if ::config(dumpinitialfile) does not exists, find the first directory that exists
	set pathname [file dirname $::config(dumpinitialfile)]
	set filename [file tail    $::config(dumpinitialfile)]
	while {![file exists $pathname]} {
		set newdir [file dirname $pathname]
		if {$pathname == $newdir} {
			set pathname $rootdir
		} else {
			set pathname $newdir
		}
	}
	set ::config(dumpinitialfile) [file join $pathname $filename]

	########################################## Tools
	set ::config(tools) {}

	# Overwrite default values with user defined ones
	set configfile [file join $::config(config_directory) "pgcedit.tools"]
	if {[file exists $configfile]} {
		if {[catch {source $configfile} err] != 0} {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}

	# set the "associate PgcEdit with .XXX" menu flags
	# based upon current registry associations
	# Only used if PgcEdit is freeWrapped and run under Windows
	# The catchs are supposed to avoid a bug in ::freewrap::shell_getCmd_imp
	if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname]} {
		if {[catch {set ::config(shortcut_startmenu) \
				[::winshellext::is_shortcut_to_PgcEdit startmenu]}]} {
			set ::config(shortcut_startmenu) 0
		}
		if {[catch {set ::config(shortcut_sendto) \
				[::winshellext::is_shortcut_to_PgcEdit sendto]}]} {
			set ::config(shortcut_sendto) 0
		}
		if {[catch {set ::config(shortcut_desktop) \
				[::winshellext::is_shortcut_to_PgcEdit desktop]}]} {
			set ::config(shortcut_desktop) 0
		}
		if {[catch {set ::config(assoicated_with_IFO) \
				[::winshellext::shell_is_associated ".IFO" "PgcEdit.exe"]}]} {
			set ::config(assoicated_with_IFO) 0
		}
		if {[catch {set ::config(assoicated_with_BUP) \
				[::winshellext::shell_is_associated ".BUP" "PgcEdit.exe"]}]} {
			set ::config(assoicated_with_BUP) 0
		}
	}
	# minimum border width for some GUI elements (checkbuttons...)
	if {$::config(tk) == "win" || $::config(tk) == "aqua"} {
		set ::config(min_bd) 0
	} else {
		set ::config(min_bd) 2
	}
}

#############################################################################
## Procedure:  save_user_prefs
## save setups
proc save_user_prefs {} {
	set txt "# PgcEdit config file.  Saved automatically.  Do not edit!\n"
	set txt "${txt}\nset ::config(global_font) [list $::config(global_font)]"
	set txt "${txt}\nset ::config(fixed_width_font) [list $::config(fixed_width_font)]"
	set txt "${txt}\nset ::config(tearoff_menus) $::config(tearoff_menus)"
	set txt "${txt}\nset ::config(save_setups_on_exit) $::config(save_setups_on_exit)"
	set txt "${txt}\nset ::config(use_file_browser) $::config(use_file_browser)"
	set txt "${txt}\nset ::config(auto_open_new_dvd) $::config(auto_open_new_dvd)"
	set txt "${txt}\nset ::config(open_trace_mode) $::config(open_trace_mode)"
	set txt "${txt}\nset ::config(auto_create_backup) $::config(auto_create_backup)"
	set txt "${txt}\nset ::config(auto_fix_streams) $::config(auto_fix_streams)"
	set txt "${txt}\nset ::config(backup_in_root) $::config(backup_in_root)"
	set txt "${txt}\nset ::config(save_BUP_with_gaps) $::config(save_BUP_with_gaps)"
	set txt "${txt}\nset ::config(selector) $::config(selector)"
	set txt "${txt}\nset ::config(selector_side) $::config(selector_side)"
	set txt "${txt}\nset ::config(selector_height) $::config(selector_height)"
	set txt "${txt}\nset ::config(buttons) $::config(buttons)"
	set txt "${txt}\nset ::config(toolbar) $::config(toolbar)"
	set txt "${txt}\nset ::config(searchstring) [list $::config(searchstring)]"
	set txt "${txt}\nset ::config(adjust_win_size) $::config(adjust_win_size)"
	set txt "${txt}\nset ::config(use_custom_color) $::config(use_custom_color)"
	set txt "${txt}\nset ::config(custom_color) $::config(custom_color)"
	set txt "${txt}\nset ::config(white_listbox_background) $::config(white_listbox_background)"
	set txt "${txt}\nset ::config(highlight_gotos) $::config(highlight_gotos)"
	set txt "${txt}\nset ::config(show_hex_codes) $::config(show_hex_codes)"
	set txt "${txt}\nset ::config(dump_hex_codes) $::config(dump_hex_codes)"
	set txt "${txt}\nset ::config(allow_auto_comments) $::config(allow_auto_comments)"
	set txt "${txt}\nset ::config(auto_backup_comments) $::config(auto_backup_comments)"
	set txt "${txt}\nset ::config(renumber_gotos) $::config(renumber_gotos)"
	set txt "${txt}\nset ::config(popups_button) $::config(popups_button)"
	set txt "${txt}\nset ::config(expert_mode) $::config(expert_mode)"
	set txt "${txt}\nset ::config(auto_launch_menubuttons_parser) $::config(auto_launch_menubuttons_parser)"
	set txt "${txt}\nset ::config(lbas_to_skip) $::config(lbas_to_skip)"
	set txt "${txt}\nset ::config(check_multiple_BOVs_sets) $::config(check_multiple_BOVs_sets)"
	set txt "${txt}\nset ::config(initialdir) [list [file nativename $::config(initialdir)]]"
	set txt "${txt}\nset ::config(importmenuinitialdir) [list [file nativename $::config(importmenuinitialdir)]]"
	set txt "${txt}\nset ::config(scratchbook_initialfile) [list [file nativename $::config(scratchbook_initialfile)]]"
	set txt "${txt}\nset ::config(import_initialdir) [list [file nativename $::config(import_initialdir)]]"
	set txt "${txt}\nset ::config(dumpinitialfile) [list [file nativename $::config(dumpinitialfile)]]"
	if {[winfo width .]==[winfo screenwidth .] && [winfo rootx .]==0} {
		set txt "${txt}\nset ::config(main_win_geometry) $::config(main_win_geometry)"
	} else {
		set txt "${txt}\nset ::config(main_win_geometry) [wm geometry .]"
	}
	if {[winfo exists .cmdeditor]} {
		set txt "${txt}\nset ::config(ced_win_geometry) [wm geometry .cmdeditor]"
	} else {
		set txt "${txt}\nset ::config(ced_win_geometry) \"$::config(ced_win_geometry)\""
	}
	if {[winfo exists .trinfo]} {
		set txt "${txt}\nset ::config(trace_info_win_geometry) [wm geometry .trinfo]"
	} else {
		set txt "${txt}\nset ::config(trace_info_win_geometry) \"$::config(trace_info_win_geometry)\""
	}

	if {[info exists ::menubuttons::win_geometry]} {
		set txt "${txt}\nset ::config(menubuttons_win_geometry) \"$::menubuttons::win_geometry\""
	} else {
		set txt "${txt}\nset ::config(menubuttons_win_geometry) \"$::config(menubuttons_win_geometry)\""
	}
	set txt "${txt}\nset ::config(menueditor_win_geometry) \"$::config(menueditor_win_geometry)\""
	set txt "${txt}\nset ::config(colorschemes_win_geometry) \"$::config(colorschemes_win_geometry)\""

	set txt "${txt}\nset ::config(use_preview) $::config(use_preview)"
	set txt "${txt}\nset ::config(preview_screen_divisor) $::config(preview_screen_divisor)"
	set txt "${txt}\nset ::tr::hide_gui $::tr::hide_gui"
	set txt "${txt}\nset ::tr::watch_always_on_top [list $::tr::watch_always_on_top]"
	set txt "${txt}\nset ::tr::highlight_log_step $::tr::highlight_log_step"
	set txt "${txt}\nset ::tr::view_gprms $::tr::view_gprms"
	set txt "${txt}\nset ::tr::view_sprms $::tr::view_sprms"
	set txt "${txt}\nset ::tr::view_puos  $::tr::view_puos"
	set txt "${txt}\nset ::tr::view_input $::tr::view_input"
	set txt "${txt}\nset ::tr::view_slabs $::tr::view_slabs"

	set txt "${txt}\nset ::config(preview_autoplay) $::config(preview_autoplay)"
	if {$::tcl_platform(platform) == "windows"} {
		set txt "${txt}\nset ::config(preview_no_overlay) $::config(preview_no_overlay)"
	}
	set txt "${txt}\nset ::config(trace_auto_preview) $::config(trace_auto_preview)"
	set txt "${txt}\nset ::config(trace_skip_small_cells_preview) $::config(trace_skip_small_cells_preview)"

	set txt "${txt}\nset ::tr::delay $::tr::delay"
	set txt "${txt}\nset ::config(trace_auto_close) $::config(trace_auto_close)"

	set txt "${txt}\nset ::config(menued_outline_color) $::config(menued_outline_color)"
	set txt "${txt}\nset ::config(menued_phantom_color) $::config(menued_phantom_color)"
	set txt "${txt}\nset ::config(menued_other_buttons_color) $::config(menued_other_buttons_color)"
	set txt "${txt}\nset ::config(menued_show_all_buttons) $::config(menued_show_all_buttons)"
	set txt "${txt}\nset ::config(menued_show_links) $::config(menued_show_links)"
	set txt "${txt}\nset ::config(menued_advanced_controls) $::config(menued_advanced_controls)"

	set txt "${txt}\nset ::config(web_browser) [list $::config(web_browser)]"

	# Burn function parameters
	if {$::tcl_platform(platform) != "macintosh"} {
		set txt "${txt}\nset ::burn::showlog $::burn::showlog"
		set txt "${txt}\nset ::burn::out [list $::burn::out]"
		set txt "${txt}\nset ::burn::mount $::burn::mount"
		set txt "${txt}\nset ::burn::mount_device $::burn::mount_device"
		set txt "${txt}\nset ::burn::burn $::burn::burn"
		set txt "${txt}\nset ::burn::speed $::burn::speed"
		set txt "${txt}\nset ::burn::media [list $::burn::media]"
		set txt "${txt}\nset ::burn::eraserw [list $::burn::eraserw]"
		set txt "${txt}\nset ::burn::verify $::burn::verify"
		set txt "${txt}\nset ::burn::start [list $::burn::start]"
		set txt "${txt}\nset ::burn::close [list $::burn::close]"
		set txt "${txt}\nset ::burn::shutdown [list $::burn::shutdown]"
		set txt "${txt}\nset ::burn::deleteimage [list $::burn::deleteimage]"
	}

	set txt "${txt}\n\n# Setups not availables in the GUI:"

	set txt "${txt}\n\n# Number of sectors for AutoLocate function of Preview"
	set txt "${txt}\nset ::config(preview_autolocate) $::config(preview_autolocate)"

	if {$::config(debug) > 0} {
		set txt "${txt}\n\nset ::config(debug) $::config(debug)"
	}
	set txt "${txt}\n"

	saveFile [file join $::config(config_directory) "pgcedit.cfg"] auto $txt
}

#############################################################################
## Procedure:  cleanexit
## save setups and exit
proc cleanexit {code} {
	if {$::config(use_preview)} {
		::preview::close_preview
	}
	if {$::dvd_modified && $code == 0} {
		set r [tk_messageBox -message "The DVD has been modified.\n\nSave it before quitting?" \
				-icon question -title "PgcEdit" -type yesnocancel -default yes]
		if {$r == "cancel"} return
		if {$r == "yes"} {
			set rc [::main::Write_DVD false]
			if {$rc != 0} return
		}
	} elseif {[info exists ::pgc_comments_modified] && $::pgc_comments_modified && !$::config(expert_mode)} {
		if {$::config(auto_backup_comments)} {
			set r [tk_messageBox -title "PgcEdit - Quit" -icon question -parent . -type okcancel -default ok \
					-message "User's PGC comments changed.\nSave them in backup folder now?"]
			if {$r == "ok"} {
				::utils::export_pgc_comments true
			}
		} else {
			::utils::export_pgc_comments false
		}
	}
	if {$::config(save_setups_on_exit) && $::config(restore_defaults)==false} {
		save_user_prefs
	}
	exit $code
}

#############################################################################
## Procedure:  errmsg
## Display an error message
proc errmsg {errnum msg1 msg2} {
	# Display a message
	puts "Error $errnum: $msg1"
	puts $msg2
	if {[info exists errorInfo]} {catch {puts $errorInfo}}
	if {$::config(debug) && $::tcl_platform(platform) == "windows"} {
		console show
	}
	show_warning -icon error -title "PgcEdit $::program_version error" \
		-message "Error $errnum $msg1\n$msg2" -parent .
}

#############################################################################
## Procedure:  readFile
## read a file and return the datas
## return "" in case of error
proc readFile {filename translation} {
	set errnum [catch {set in [open $filename r]} err]
	if {$errnum} {
		errmsg $errnum "opening file for read" $err
		return ""
	}
	fconfigure $in -translation $translation
	set errnum [catch {set data [read $in]} err]
	if {$errnum} {
		close $in
		errmsg $errnum "reading file" $err
		return ""
	}
	close $in
	return $data
}

#############################################################################
## Procedure:  saveFile
## save datas in a file
## return error number, or 0 for OK
## if $args == "-append", then append to end of an existing file, else overwrite
proc saveFile {filename translation data args} {
	if {$args == "-append"} {
		set mode a
		set modetext "append"
	} else {
		set mode w
		set modetext "write"
	}
	set errnum [catch {set out [open $filename $mode]} err]
	if {$errnum} {
		errmsg $errnum "opening file for $modetext" $err
		return $errnum
	}
	fconfigure $out -translation $translation
	if {$translation == "binary"} {
		set errnum [catch {puts -nonewline $out $data} err]
	} else {
		set errnum [catch {puts $out $data} err]
	}
	if {$errnum} {
		errmsg $errnum "writing file" $err
	}
	close $out
	return $errnum
}

#############################################################################
## Simple interfaces for file handling, with 'Retry/Cancel' requester in case of error.
## return true if a non-recoverable error occured.
proc movefile {source destination {force false} {title PgcEdit} {parent .}} {	
	set rc 0
	set done false
	while {!$done} {
		if {$force=="1" || $force=="true" || $force=="-force"} {
			set rc [catch {file rename -force -- $source $destination} errstr]
		} else {
			set rc [catch {file rename -- $source $destination} errstr]
		}
		set done true
		if {$rc} {
			set r [tk_messageBox -title $title -icon warning -parent $parent -type retrycancel -default retry \
					-message "$errstr"]
			if {$r == "retry"} {
				set done false
			}
		}
	}
	return $rc
}
proc copyfile {source destination force {title PgcEdit} {parent .}} {	
	set rc 0
	set done false
	while {!$done} {
		if {$force=="1" || $force=="true" || $force=="-force"} {
			set rc [catch {file copy -force -- $source $destination} errstr]
		} else {
			set rc [catch {file copy -- $source $destination} errstr]
		}
		set done true
		if {$rc} {
			set r [tk_messageBox -title $title -icon warning -parent $parent -type retrycancel -default retry \
					-message "$errstr"]
			if {$r == "retry"} {
				set done false
			}
		}
	}
	return $rc
}
proc deletefile {file {title PgcEdit} {parent .}} {
	if {![file exists $file]} {
		return 0
	}
	set rc 0
	set done false
	while {!$done} {
		set rc [catch {file delete -force -- $file} errstr]
		set done true
		if {$rc} {
			set r [tk_messageBox -title $title -icon warning -parent $parent -type retrycancel -default retry \
					-message "$errstr"]
			if {$r == "retry"} {
				set done false
			}
		}
	}
	return $rc
}
proc makedir {dir {title PgcEdit} {parent .}} {	
	if {[file exists $dir] && [file isdirectory $dir]} {
		return 0
	}
	set rc 0
	set done false
	while {!$done} {
		set rc [catch {file mkdir $dir} errstr]
		set done true
		if {$rc} {
			set r [tk_messageBox -title $title -icon warning -parent $parent -type retrycancel -default retry \
					-message "$errstr"]
			if {$r == "retry"} {
				set done false
			}
		}
	}
	return $rc
}

#############################################################################
## Procedure:  b2x
## Convert a binary string to hexadecimal
proc b2x {binarystring} {
	binary scan $binarystring H* hex
	return $hex
}

#############################################################################
## Procedure:  b2i
## Convert a binary string to decimal
proc b2i {binarystring} {
	binary scan $binarystring H* hex
	return [x2d $hex]
}

#############################################################################
## Procedure:  d2c
## Convert a value to binary characters
proc d2c {val numbytes} {
	switch $numbytes {
		1 {return [binary format c $val]}
		2 {return [binary format S $val]}
		4 {return [binary format I $val]}
		8 {return [binary format W $val]}
	}
}

#############################################################################
## Procedure:  x2d
## Convert an hexadecimal string to decimal value
## IMPORTANT: Do not thrust the TclChecker here: leave this proc as it is!
proc x2d {h} {return [expr 0x$h]}

#############################################################################
## Procedures:  min and max
proc max {a b} {
	if {$a>$b} {return $a}
	return $b
}
proc min {a b} {
	if {$a<$b} {return $a}
	return $b
}

#############################################################################
## Procedure:  get_int
## Get an integer from inside the binary data
proc get_int {binarystring_upvar start_offset length} {
	upvar $binarystring_upvar binarystring
	set s [string range $binarystring $start_offset [expr {$start_offset+$length-1}]]
	binary scan $s H* hex
	if {$hex!=""} {return [x2d $hex]}

	set txt "Fatal error!\n"
	set txt "$txt\nPgcEdit is trying to read data after the"
	set txt "$txt\nend of a table."
	set txt "$txt\nThis means that a Length Of Table value"
	set txt "$txt\nis probably wrong.\n"
	set txt "$txt\nPlease try to fix the problem with IfoEdit.\n"
	set txt "$txt\nThe missing value may be fixed to 0, but"
	set txt "$txt\nthis could lead to strange results.\n"
	set txt "$txt\nDo you want to fix the value and continue"
	set txt "$txt\nanyway? (NOT RECOMMENDED!)"
	set r [tk_messageBox -title "PgcEdit fatal error" -icon error -message $txt \
			-parent . -type yesno -default yes]
	if {$r == "no"} {error "Attempt to read data after end of table"}
	set hex 00
	return [x2d $hex]
}

#############################################################################
## Procedure:  set_int
## Set an integer inside the binary data whose variable is passed in arguments
proc set_int {binary_upvar start_offset value length} {
	upvar $binary_upvar binarystring
	set binarystring [string replace $binarystring $start_offset [expr {$start_offset+$length-1}] [d2c $value $length]]
	return
}

#############################################################################
# return the number of pre, post or cell commands in the provided commands
proc num_type_cmds {type cmds_upvar} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {return [expr {[lindex $idxs 1]-[lindex $idxs 0]-1}]}
		post {return [expr {[lindex $idxs 2]-[lindex $idxs 1]-1}]}
		cell {return [expr {[llength $cmds]-[lindex $idxs 2]-1}]}
	}
}

#############################################################################
# return the pre, post or cell commands from the provided commands
proc get_type_cmds {type cmds_upvar} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {return [lrange $cmds [expr {[lindex $idxs 0]+1}] [expr {[lindex $idxs 1]-1}]]}
		post {return [lrange $cmds [expr {[lindex $idxs 1]+1}] [expr {[lindex $idxs 2]-1}]]}
		cell {return [lrange $cmds [expr {[lindex $idxs 2]+1}] end]}
	}
}

#############################################################################
# return the type ("pre", "post" or "cell") of the Nth. command in the provided commands.
# If the cmd number is a pre, post or cell label, return "* pre", "* post" or "* cell".
proc get_cmd_type {cmds_upvar index} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	if {[llength $idxs] == 1} {
		if {$index==[lindex $idxs 0]} {return "* scratchbook"}
		return "scratchbook"
	}
	if {$index==[lindex $idxs 0]} {return "* pre"}
	if {$index==[lindex $idxs 1]} {return "* post"}
	if {$index==[lindex $idxs 2]} {return "* cell"}
	if {$index<[lindex $idxs 1]} {return "pre"}
	if {$index<[lindex $idxs 2]} {return "post"}
	return "cell"
}

#############################################################################
# return the line number of the command into his type section.
# If the cmd number is a pre, post or cell label, return 0.
proc get_cmd_linenum {cmds_upvar index} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	if {[llength $idxs] == 1} {
		return $index	;# index within scratchbook commands
	}
	if {$index<[lindex $idxs 1]} {return $index}
	if {$index<[lindex $idxs 2]} {return [expr {$index - [lindex $idxs 1]}]}
	return [expr {$index - [lindex $idxs 2]}]
}

#############################################################################
# replace the pre, post or cell commands in the commands list by the provided commands
proc set_type_cmds {type cmds_upvar newcmds} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {
			set idx1 [lindex $idxs 0]
			set idx2 [lindex $idxs 1]
		}
		post {
			set idx1 [lindex $idxs 1]
			set idx2 [lindex $idxs 2]
		}
		cell {
			set idx1 [lindex $idxs 2]
			set idx2 [llength $cmds]
		}
	}
	set cmds [concat [lrange $cmds 0 $idx1] $newcmds [lrange $cmds $idx2 end]]
}

#############################################################################
# refresh the listbox with the current PGC
proc refresh_current_pgc {} {
if {$::scratchbook_active} {
		return
	}
	set selectionindexes [::main::get_selection_indexes]
	set type [lindex $::editor_cmds_from 0]
	set pgcindex [split [lindex $::editor_cmds_from 1] ","]
	set vts [lindex $pgcindex 0]
	switch $type {
		pgc {
			::main::display_pgc $type $vts [lindex $pgcindex 1]
		}
		menu {
			::main::display_pgc $type $vts [lindex $pgcindex 1] [lindex $pgcindex 2]
		}
	}
	::main::set_listbox_selection [lindex $selectionindexes 0] [lindex $selectionindexes 1]
}

#############################################################################
## PROGRAM MAIN ENTRY
#############################################################################

# Parse arguments
if {$argc == 1 && ([lindex $argv 0] == "/?" || [lindex $argv 0] == "-h")} {
	puts "PgcEdit v $::program_version"
	puts "Command line usage:"
	puts "> PgcEdit \[\"DVD pathname\"\]"
	puts "or"
	puts "> PgcEdit \[\"DVD pathname\\any file\"\]"
	puts "The second form is used to associate PgcEdit with IFO (or BUP) files."
	exit 0
}

set ::config(tempdir) "NOT_ASSIGNED"
catch {set ::config(tempdir) $::env(TEMP)}
if {! [file isdirectory $::config(tempdir)]} {
	catch {set ::config(tempdir) $::env(TMP)}
}
if {! [file isdirectory $::config(tempdir)]} {
	if {[file isdirectory "/tmp"] && [file writable "/tmp"]} {
		catch {set ::config(tempdir) "/tmp"}
	}
}

# init some system defaults
# we need to know some widget default colors
set ::config(illegal_aps) false
if {$::tcl_platform(platform) == "windows"} {
	set ::config(preview_available) 1
	set ::config(use_preview) true
	set ::config(preview_launcher) ""
	if {! [catch {set geocode [registry get "HKEY_CURRENT_USER\\Control Panel\\International\\Geo" "Nation"]}]} {
		if {[lsearch {94 242} $geocode] != -1} {
			set ::config(illegal_aps) true
		}
	} else {
		if {! [catch {set geocode [registry get "HKEY_CURRENT_USER\\Control Panel\\International" "sCountry"]}]} {
			if {[lsearch {"Germany" "United Kingdom"} $geocode] != -1} {
				set ::config(illegal_aps) true
			} 
		}
	}
} else {
	set ::config(preview_available) 0
	set ::config(use_preview) false
	if {$::tcl_platform(platform) == "unix"} {
		if {![catch {exec -- wine --version 2>/dev/null} errstr]} {
			set ::config(preview_available) 1
			set ::config(use_preview) true
			set ::config(preview_launcher) "wine"
		}
		if {![catch {exec -- darwine --version 2>/dev/null} errstr]} {
			set ::config(preview_available) 1
			set ::config(use_preview) true
			set ::config(preview_launcher) "darwine"
		}
	}
}

# Full path is needed for FreeWrap
set scriptname [info script]
if {[info exists ::freewrap::progname]} {
	# FreeWrap as wish or Wrapped script
	if {$scriptname == ""} {
		# Wrapped script
		set config(apppath) "/Tcl/work/PGCEDIT"
	} else {
		# FreeWrap as wish
		set config(apppath) [file dirname $scriptname]
	}
	if {$::tcl_platform(platform) == "windows"} {
		source "$config(apppath)/lib/winshellext.tcl"
	}
	set ::config(bindir) [file join [file dirname $argv0] "bin"]
	set ::config(plugindir) [file join [file dirname $argv0] "plugins"]
	if {! [file exists $::config(tempdir)]} {
		makedir [file join [file dirname $argv0] "temp"]
		set ::config(tempdir) [file join [file dirname $argv0] "temp"]
	}

	if {$::tcl_platform(platform) == "windows" || $::config(preview_available)} {
		set copy 0
		if {! [file exists [file join $::config(bindir) "binversion.tcl"]]} {
			set copy 1
		} else {
			source [file join $::config(bindir) "binversion.tcl"]
			if {$bin_version != $::program_version} {
				set copy 1
			}
		}
		if {$copy} {
			puts "Installing the binaries v$::program_version in \"[file nativename $::config(bindir)]\"."
			puts "You need write permission on this folder and his parent to complete the install."
			puts "Also, be sure to close all instances of PgcEdit before launching a new version."
			catch {file attributes [file dirname $::config(bindir)] -readonly false}
			makedir $::config(bindir)
			catch {file attributes $::config(bindir) -readonly false}
			saveFile [file join $::config(bindir) "PgcEditPreview.exe"] binary [readFile [file join $config(apppath) "bin/PgcEditPreview.exe"] binary]
			catch {file attributes [file join $::config(bindir) "PgcEditPreview.exe"] -readonly false}
			if {$::tcl_platform(platform) == "windows"} {
				saveFile [file join $::config(bindir) "pskill.exe"] binary [readFile [file join $config(apppath) "bin/pskill.exe"] binary]
				catch {file attributes [file join $::config(bindir) "pskill.exe"] -readonly false}
			}
			catch {file delete [file join $::config(bindir) "mfc42.dll"]}	;# not needed now
			catch {file delete [file join $::config(bindir) "pkgIndex.tcl"]}	;# not needed now
			catch {file delete [file join $::config(bindir) "tkimg13.dll"]}	;# not needed now
			catch {file delete [file join $::config(bindir) "tkimgbmp13.dll"]}	;# not needed now
			catch {file delete [file join $::config(bindir) "BlankCell.exe"]}	;# not needed now
			catch {file delete [file join $::config(bindir) "dvd2bmp.exe"]}	;# not needed now
			catch {file delete [file join $::config(bindir) "kill.exe"]}	;# not needed now
			if {[file exists [file join $::config(bindir) "cygwin1.dll"]]} {
				catch {file delete [file join $::config(bindir) "cygwin1.dll"]}
			}
			if {[file exists [file join $::config(bindir) "button_dump.exe"]]} {
				catch {file delete [file join $::config(bindir) "button_dump.exe"]}
			}
			saveFile [file join $::config(bindir) "binversion.tcl"] auto "set bin_version $::program_version\n"
			catch {file attributes [file join $::config(bindir) "binversion.tcl"] -readonly false}
			puts "Binaries v$::program_version copied successfully."
		}
	} else {
		if {[file exists [file join $::config(bindir) "button_dump_osx"]]} {
			catch {file delete [file join $::config(bindir) "button_dump_osx"]}
		}
		if {[file exists [file join $::config(bindir) "button_dump_linux"]]} {
			catch {file delete [file join $::config(bindir) "button_dump_linux"]}
		}
		if {[file exists [file join $::config(bindir) "binversion.tcl"]]} {
			catch {file delete [file join $::config(bindir) "binversion.tcl"]}
		}
		catch {file delete $::config(bindir)}
	}
} else {
	set config(apppath) [file dirname $scriptname]
	set ::config(bindir) [file join [file dirname $scriptname] "bin"]
	set ::config(plugindir) [file join [file dirname $scriptname] "plugins"]

	if {! [file exists $::config(tempdir)]} {
		set ::config(tempdir) "/tmp"
	}
	if {! [file exists $::config(tempdir)]} {
		makedir [file join [file dirname $scriptname] "temp"]
		set ::config(tempdir) [file join [file dirname $scriptname] "temp"]
	}

	if {$::config(preview_available)} {
		source [file join $::config(bindir) "binversion.tcl"]
		if {$bin_version != $::program_version} {
			tk_messageBox -icon error -title "PgcEdit $::program_version error" -parent . \
				-message "The binaries (in bin subdirectory)\nare either missing or obsolete!\nPlease reinstall PgcEdit."
		}
	}
	if {[file exists [file join $::config(bindir) "cygwin1.dll"]]} {
		catch {file delete [file join $::config(bindir) "cygwin1.dll"]}
	}
	if {[file exists [file join $::config(bindir) "button_dump.exe"]]} {
		catch {file delete [file join $::config(bindir) "button_dump.exe"]}
	}
	if {[file exists [file join $::config(bindir) "button_dump_osx"]]} {
		catch {file delete [file join $::config(bindir) "button_dump_osx"]}
	}
	if {[file exists [file join $::config(bindir) "button_dump_linux"]]} {
		catch {file delete [file join $::config(bindir) "button_dump_linux"]}
	}
}
if {! [file exists $::config(plugindir)]} {
	catch {makedir $::config(plugindir)}
}

# init user preferences
init_user_prefs
if {! $::config(preview_available)} {
	set ::config(use_preview) false
}
button .b
set ::config(default_color) [.b cget -background]
if {$::config(use_custom_color)} {
	tk_setPalette $::config(custom_color)
}

# init some system defaults
# we need to know some widget default colors
set config(systembuttonface) [.b cget -background]
set config(disabledforeground) [.b cget -disabledforeground]
if {$::tcl_platform(platform) == "windows"} {
	set ::config(systemhighlight) systemhighlight
} else {
	set ::config(systemhighlight) gray75
}
destroy .b

# disable the dprint routine if it is not needed, for speed purposes
if {$::config(debug) == 0} {
	proc dprint {args} {}
}

if {$::config(white_listbox_background)} {
	set ::config(system_window_color) white
} else {
	set ::config(system_window_color) [[entry .e] cget -background]
	destroy .e
}
option add *Listbox*background $::config(system_window_color)
option add *Entry*background $::config(system_window_color)
option add *font $::config(global_font)
option add *tearOff $::config(tearoff_menus)
option add *Scale*troughColor $::config(system_window_color)
option add *Scale*sliderLength 15
option add *Scale*highlightThickness 1
option add *Scale*borderWidth 1
option add *Scale*orient horizontal
if {$::config(tk) == "aqua"} {
	option add *Menubutton*borderWidth 0
	option add *Button*padX 11
	option add *Entry*borderWidth 1
	option add *Entry*highlightThickness 1
} else {
	option add *Menubutton*relief raised
	option add *Menubutton*highlightThickness 1
	option add *Menubutton*indicatorOn true
	option add *Button*padX 1
}
option add *Menubutton*takeFocus _focus_menubutton
proc _focus_menubutton {w} {if {[$w cget -state]=="disabled"} {return 0} else {return 1}}

if {$::config(tk) != "x11"} {
	option add *Radiobutton*selectColor $::config(system_window_color)
	option add *Checkbutton*selectColor $::config(system_window_color)
} else {
	option add *Radiobutton*borderWidth 1
	option add *Scrollbar*width 11
	#option add *Scrollbar*troughColor #E0E0E0	
}


# read program modules
source $config(apppath)/lib/tk_menu_bugfix.tcl	;# MenuUnpost Tk bug fixed by r0lZ
source $config(apppath)/lib/readdvd.tcl
source $config(apppath)/lib/menued.tcl
source $config(apppath)/lib/writedvd.tcl
source $config(apppath)/lib/disasm.tcl
source $config(apppath)/lib/utils.tcl
source $config(apppath)/lib/macros.tcl
source $config(apppath)/lib/toolscfg.tcl
source $config(apppath)/lib/cmdedit.tcl
source $config(apppath)/lib/pgceditor.tcl
source $config(apppath)/lib/trace.tcl
source $config(apppath)/lib/srptinfo.tcl
source $config(apppath)/lib/streams.tcl
source $config(apppath)/lib/menubuttons.tcl
source $config(apppath)/lib/blankcell.tcl
if {$::config(preview_available)} {
	source $config(apppath)/lib/preview.tcl
}
source $config(apppath)/lib/toolbar.tcl
source $config(apppath)/lib/stringreq.tcl
source $config(apppath)/lib/fontreq.tcl
source $config(apppath)/lib/calculator.tcl
source $config(apppath)/lib/burn.tcl
source $config(apppath)/lib/menus.tcl
source $config(apppath)/lib/main.tcl			;# main program starts here

# for Tcl/Tk programmers: add your private functions in file ~/.pgcedit.user
set userfile [file join $::config(config_directory) "pgcedit.user.tcl"]
if {[file exists $userfile]} {source $userfile}
unset userfile

# Load plugins, if any
if {[file isdirectory $::config(plugindir)]} {
	set filelist [lsort [glob -nocomplain -directory $::config(plugindir) "*.tcl"]]
	if {[llength $filelist] > 0} {
		if {! [winfo exists .mainmenu.plugins]} {
			.mainmenu insert Tools cascade -label Plugins -menu .mainmenu.plugins
			menu .mainmenu.plugins
		}
		foreach f $filelist {
			source $f
		}
		unset f
	}
	unset filelist
}
